// SPDX-License-Identifier: MIT
pragma solidity >=0.8.0 <0.9.0;

import "@openzeppelin/contracts-upgradeable/proxy/utils/Initializable.sol";
import "@openzeppelin/contracts-upgradeable/security/PausableUpgradeable.sol";
import "@openzeppelin/contracts-upgradeable/access/AccessControlUpgradeable.sol";
import "@openzeppelin/contracts-upgradeable/proxy/utils/UUPSUpgradeable.sol";
import "@openzeppelin/contracts-upgradeable/security/ReentrancyGuardUpgradeable.sol";
import "@openzeppelin/contracts/utils/cryptography/ECDSA.sol";
import "@openzeppelin/contracts/utils/Context.sol";
import "./IGateway.sol";
import "./SignatureUtils.sol";
import "./IApplication.sol";
import "./ICrossTalkApplication.sol";
import "./Utils.sol";

// import "hardhat/console.sol";
/**
    @title Facilitates request to Router, validator set update and request from router.
    @author Router Protocol
 */
contract GatewayUpgradeable is
    Initializable,
    PausableUpgradeable,
    AccessControlUpgradeable,
    UUPSUpgradeable,
    ReentrancyGuardUpgradeable
{
    // storage
    string public chainId;
    uint64 public chainType;
    // nonce for requestToRouter Messages
    uint64 public eventNonce;
    uint64 public stateLastValsetNonce;
    // last validated ValSet Checkpoint
    bytes32 public stateLastValsetCheckpoint;
    // default fee for "requestToRouter"
    uint256 public requestToRouterDefaultFee;
    // default fee for "requestToDest"
    uint256 public requestToDestDefaultFee;

    bytes32 byteChainId;
    // bytes32 encoding of the string "checkpoint"
    bytes32 constant CHECK_POINT_METHOD_NAME = 0x636865636b706f696e7400000000000000000000000000000000000000000000;
    // bytes32 encoding of "requestFromRouter"
    bytes32 constant REQUEST_FROM_ROUTER_METHOD_NAME =
        0x7265717565737446726f6d526f75746572000000000000000000000000000000;
    // bytes32 encoding of "requestFromSource"
    bytes32 constant REQUEST_FROM_SOURCE_METHOD_NAME =
        0x7265717565737446726f6d536f75726365000000000000000000000000000000;
    // bytes32 encoding of "crossTalkAck"
    bytes32 constant CROSS_TALK_ACK_METHOD_NAME = 0x63726F737354616C6B41636B0000000000000000000000000000000000000000;

    mapping(string => mapping(uint64 => uint8)) public outboundNonceMapping;
    // cross-talk sequencial nonce mapping
    // (destChainid{string} => (destChainType{uint64} => nonce{uint64}))
    mapping(string => mapping(uint64 => uint64)) public destChainNonces;
    // cross-talk nonce validation mapping for incoming requests from different chains
    // (srcChainid{string} => (srcChainType{uint64} => (nonce{uint64} => handled{uint8})))
    mapping(string => mapping(uint64 => mapping(uint64 => uint8))) public crossTalkNonceMapping;

    event RequestToRouterEvent(
        address indexed applicationContract,
        uint64 indexed eventNonce,
        uint64 srcChainType,
        string routerBridgeContract,
        string srcChainId,
        uint256 gasLimit,
        bytes feePayer,
        address txOrigin,
        bytes payload
    );

    event RequestToDestEvent(
        bytes applicationContract,
        uint64 indexed eventIdentifier,
        Utils.SourceChainParams srcChainParams,
        Utils.AckGasParams ackGasParam,
        Utils.DestinationChainParams destChainParams,
        bytes[] destContractAddresses,
        bytes[] payloads,
        Utils.AckType ackType,
        bytes feePayer,
        address txOrigin,
        bool isReadCall
    );

    event ValsetUpdatedEvent(
        uint64 indexed _newValsetNonce,
        uint64 indexed _eventNonce,
        uint64 srcChainType,
        string srcChainId,
        address[] _validators,
        uint64[] _powers
    );

    // This is OutBound Request Acknowledgement event
    event EventOutboundAck(
        string relayerRouterAddress,
        string outboundTxRequestedBy,
        string chainId,
        uint64 chainType,
        uint64 indexed eventNonce,
        uint64 outboundTxNonce,
        bytes contractAckResponses,
        uint8 exeCode,
        bool status
    );

    // This is CrossTalk Request Acknowledgement event
    event CrossTalkAckEvent(
        string relayerRouterAddress,
        bytes srcContractAddress,
        string srcChainId,
        uint64 srcChainType,
        string chainId,
        uint64 chainType,
        uint64 eventIdentifier,
        uint64 crossTalkNonce,
        uint64 indexed eventNonce,
        bytes contractAckResponses,
        uint8 exeCode,
        bool status
    );

    // This is CrossTalk Acknowledgement Request Submit Event
    event CrossTalkAckReceivedEvent(
        string relayerRouterAddress,
        uint64 indexed eventNonce,
        uint64 indexed eventIdentifier,
        uint64 chainType,
        string chainId,
        bool success,
        bytes data
    );

    error ExecuteCallsRevert(bool[] execFlags, bytes[] execData);

    function validateValsetPower(uint64[] memory _powers) private pure {
        uint64 cumulativePower = 0;
        for (uint64 i = 0; i < _powers.length; i++) {
            cumulativePower = cumulativePower + _powers[i];
            if (cumulativePower > Utils.constantPowerThreshold) {
                break;
            }
        }
        if (cumulativePower <= Utils.constantPowerThreshold) {
            revert Utils.InsufficientPower({
                cumulativePower: cumulativePower,
                powerThreshold: Utils.constantPowerThreshold
            });
        }
    }

    modifier isSelf() {
        require(_msgSender() == address(this), "ExecuteHandleCalls : Can only be called by Current Contract");
        _;
    }

    // Make a new checkpoint from the supplied validator set
    // A checkpoint is a hash of all relevant information about the valset. This is stored by the contract,
    // instead of storing the information directly. This saves on storage and gas.
    // The format of the checkpoint is:
    // h(_chainId, "checkpoint", valsetNonce, validators[], powers[])
    // Where h is the keccak256 hash function.
    // The validator powers must be decreasing or equal. This is important for checking the signatures on the
    // next valset, since it allows the caller to stop verifying signatures once a quorum of signatures have been verified.
    function makeCheckpoint(
        Utils.ValsetArgs memory _valsetArgs,
        string memory _chainId
    ) internal pure returns (bytes32) {
        bytes32 checkpoint = keccak256(
            abi.encode(
                _chainId,
                CHECK_POINT_METHOD_NAME,
                _valsetArgs.valsetNonce,
                _valsetArgs.validators,
                _valsetArgs.powers
            )
        );

        return checkpoint;
    }

    // This updates the valset by checking that the validators in the current valset have signed off on the
    // new valset. The signatures supplied are the signatures of the current valset over the checkpoint hash
    // generated from the new valset.
    // Anyone can call this function, but they must supply valid signatures of constantPowerThreshold of the current valset over
    // the new valset.
    function updateValset(
        // The new version of the validator set
        Utils.ValsetArgs calldata _newValset,
        // The current validators that approve the change
        Utils.ValsetArgs calldata _currentValset,
        // These are arrays of the parts of the current validator's signatures
        Utils.Signature[] calldata _sigs
    ) external nonReentrant whenNotPaused {
        // CHECKS

        // Check that the valset nonce is greater than the old one
        if (_newValset.valsetNonce <= _currentValset.valsetNonce) {
            revert Utils.InvalidValsetNonce({
                newNonce: _newValset.valsetNonce,
                currentNonce: _currentValset.valsetNonce
            });
        }

        // Check that the valset nonce is less than a million nonces forward from the old one
        // this makes it difficult for an attacker to lock out the contract by getting a single
        // bad validator set through with uint256 max nonce
        if (_newValset.valsetNonce > _currentValset.valsetNonce + 1000000) {
            revert Utils.InvalidValsetNonce({
                newNonce: _newValset.valsetNonce,
                currentNonce: _currentValset.valsetNonce
            });
        }

        // Check that new validators and powers set is well-formed
        if (_newValset.validators.length != _newValset.powers.length || _newValset.validators.length == 0) {
            revert Utils.MalformedNewValidatorSet();
        }

        // Check cumulative power to ensure the contract has sufficient power to actually
        // pass a vote
        validateValsetPower(_newValset.powers);

        // Check that the supplied current validator set matches the saved checkpoint
        if (makeCheckpoint(_currentValset, chainId) != stateLastValsetCheckpoint) {
            revert Utils.IncorrectCheckpoint();
        }

        // Check that enough current validators have signed off on the new validator set
        bytes32 newCheckpoint = makeCheckpoint(_newValset, chainId);
        bytes32 digest = _make_digest(newCheckpoint);
        SignatureUtils.checkValidatorSignatures(_currentValset, _sigs, digest, Utils.constantPowerThreshold);

        // ACTIONS

        // Stored to be used next time to validate that the valset
        // supplied by the caller is correct.
        stateLastValsetCheckpoint = newCheckpoint;

        // Store new nonce
        stateLastValsetNonce = _newValset.valsetNonce;
        eventNonce++;
        emit ValsetUpdatedEvent(
            _newValset.valsetNonce,
            eventNonce,
            chainType,
            chainId,
            _newValset.validators,
            _newValset.powers
        );
    }

    /**
        @notice Initializes Bridge, creates and grants {msg.sender} the admin role,
        creates and grants {initialRelayers} the relayer role.
        @param _chainId ID of chain the Bridge contract exists on.
        @param _chainType chain type such as evm, substrate, cosmos etc.
     */
    function initialize(
        string memory _chainId,
        uint64 _chainType,
        address[] memory _validators,
        uint64[] memory _powers,
        uint64 valsetNonce
    ) external initializer {
        // CHECKS

        // Check that validators, powers, and signatures (v,r,s) set is well-formed
        if (_validators.length != _powers.length || _validators.length == 0) {
            revert Utils.MalformedCurrentValidatorSet();
        }

        // Check cumulative power to ensure the contract has sufficient power to actually
        // pass a vote
        // point -> if due to special order our set to
        validateValsetPower(_powers);

        Utils.ValsetArgs memory _valset;
        _valset = Utils.ValsetArgs(_validators, _powers, valsetNonce);

        bytes32 newCheckpoint = makeCheckpoint(_valset, _chainId);

        // ACTIONS
        __AccessControl_init();
        __Pausable_init();

        // Constructor Fx
        _setupRole(DEFAULT_ADMIN_ROLE, msg.sender);

        chainId = _chainId;
        chainType = _chainType;
        eventNonce = 0;
        stateLastValsetNonce = valsetNonce;
        stateLastValsetCheckpoint = newCheckpoint;
        byteChainId = keccak256(abi.encodePacked(_chainId));
        eventNonce++;
        emit ValsetUpdatedEvent(valsetNonce, eventNonce, chainType, chainId, _validators, _powers);
        // Constructor Fx
    }

    function _authorizeUpgrade(address newImplementation) internal virtual override onlyRole(DEFAULT_ADMIN_ROLE) {}

    function setBridgeFees(
        uint256 _requestToRouterDefaultFee,
        uint256 _requestToDestDefaultFee
    ) external onlyRole(DEFAULT_ADMIN_ROLE) {
        // default fee for "requestToRouter"
        requestToRouterDefaultFee = _requestToRouterDefaultFee;
        // default fee for "requestToDest"
        requestToDestDefaultFee = _requestToDestDefaultFee;
    }

    function requestToRouter(
        bytes memory payload,
        string memory routerBridgeContract,
        uint256 gasLimit,
        Utils.FeePayer feePayerEnum
    ) external payable whenNotPaused returns (uint64) {
        require(msg.value >= requestToRouterDefaultFee, "insufficient fees sent");
        address txOrigin;
        assembly {
            txOrigin := origin()
        }
        address feePayer = msg.sender;
        if (feePayerEnum == Utils.FeePayer.USER) {
            feePayer = txOrigin;
        } else if (feePayerEnum == Utils.FeePayer.NONE) {
            feePayer = address(0);
        }
        eventNonce++;
        emit RequestToRouterEvent(
            msg.sender,
            eventNonce,
            chainType,
            routerBridgeContract,
            chainId,
            gasLimit,
            toBytes(feePayer),
            txOrigin,
            payload
        );
        return eventNonce;
    }

    function requestToDest(
        Utils.RequestArgs memory requestArgs,
        Utils.AckType ackType,
        Utils.AckGasParams memory ackGasParams,
        Utils.DestinationChainParams memory destChainParams,
        Utils.ContractCalls memory contractCalls
    ) external payable whenNotPaused returns (uint64) {
        require(msg.value >= requestToDestDefaultFee, "insufficient fees sent");

        destChainNonces[destChainParams.destChainId][destChainParams.destChainType]++;
        // uint64 crossTalkNonce = destChainNonces[destChainParams.destChainId][destChainParams.destChainType];
        eventNonce++;
        Utils.SourceChainParams memory srcChainParams = Utils.SourceChainParams(
            destChainNonces[destChainParams.destChainId][destChainParams.destChainType],
            requestArgs.expTimestamp,
            requestArgs.isAtomicCalls,
            chainType,
            chainId
        );

        address txOrigin;
        assembly {
            txOrigin := origin()
        }
        address feePayer = msg.sender;
        if (requestArgs.feePayerEnum == Utils.FeePayer.USER) {
            feePayer = txOrigin;
        } else if (requestArgs.feePayerEnum == Utils.FeePayer.NONE) {
            feePayer = address(0);
        }
        emit RequestToDestEvent(
            toBytes(msg.sender),
            eventNonce,
            srcChainParams,
            ackGasParams,
            destChainParams,
            contractCalls.destContractAddresses,
            contractCalls.payloads,
            ackType,
            toBytes(feePayer),
            txOrigin,
            false
        );
        return eventNonce;
    }

    function readQueryToDest(
        Utils.RequestArgs memory requestArgs,
        Utils.AckGasParams memory ackGasParams,
        Utils.DestinationChainParams memory destChainParams,
        Utils.ContractCalls memory contractCalls
    ) external payable whenNotPaused returns (uint64) {
        require(msg.value >= requestToDestDefaultFee, "insufficient fees sent");

        destChainNonces[destChainParams.destChainId][destChainParams.destChainType]++;
        // uint64 crossTalkNonce = destChainNonces[destChainParams.destChainId][destChainParams.destChainType];
        eventNonce++;
        Utils.SourceChainParams memory srcChainParams = Utils.SourceChainParams(
            destChainNonces[destChainParams.destChainId][destChainParams.destChainType],
            requestArgs.expTimestamp,
            requestArgs.isAtomicCalls,
            chainType,
            chainId
        );

        address txOrigin;
        assembly {
            txOrigin := origin()
        }
        address feePayer = msg.sender;
        if (requestArgs.feePayerEnum == Utils.FeePayer.USER) {
            feePayer = txOrigin;
        } else if (requestArgs.feePayerEnum == Utils.FeePayer.NONE) {
            feePayer = address(0);
        }
        emit RequestToDestEvent(
            toBytes(msg.sender),
            eventNonce,
            srcChainParams,
            ackGasParams,
            destChainParams,
            contractCalls.destContractAddresses,
            contractCalls.payloads,
            Utils.AckType.ACK_ON_BOTH,
            toBytes(feePayer),
            txOrigin,
            true
        );
        return eventNonce;
    }

    function _make_digest(bytes32 data) private pure returns (bytes32 _digest) {
        _digest = keccak256(abi.encodePacked(Utils.MSG_PREFIX, data));
    }

    function requestFromRouter(
        // The validators that approve the call
        Utils.ValsetArgs calldata _currentValset,
        // These are arrays of the parts of the validators signatures
        Utils.Signature[] calldata _sigs,
        Utils.RouterRequestPayload memory requestPayload
    ) external nonReentrant whenNotPaused {
        // chainId, chainType data is present inside the signature validation
        // valsetArgs validation will be handled by the makeCheckPoint function
        require(
            outboundNonceMapping[requestPayload.routerBridgeAddress][requestPayload.outboundTxNonce] == 0,
            "request message already handled"
        );

        // this will validate the checkpoint, thus verify the signers
        if (makeCheckpoint(_currentValset, chainId) != stateLastValsetCheckpoint) {
            revert Utils.IncorrectCheckpoint();
        }

        bytes memory encodedABI = abi.encode(
            // bytes32 encoding of "requestFromRouter"
            REQUEST_FROM_ROUTER_METHOD_NAME,
            chainType,
            chainId,
            requestPayload.routerBridgeAddress,
            requestPayload.outboundTxNonce,
            requestPayload.isAtomic,
            requestPayload.expTimestamp,
            requestPayload.handlers,
            requestPayload.payloads
        );

        bytes32 messagehash = keccak256(encodedABI);

        bytes32 digest = _make_digest(messagehash);
        SignatureUtils.checkValidatorSignatures(_currentValset, _sigs, digest, Utils.constantPowerThreshold);

        outboundNonceMapping[requestPayload.routerBridgeAddress][requestPayload.outboundTxNonce] = 1;
        eventNonce++;
        uint64 currentNonce = eventNonce;
        if (block.timestamp > requestPayload.expTimestamp) {
            bytes memory emptyData = "";
            emit EventOutboundAck(
                requestPayload.relayerRouterAddress,
                requestPayload.routerBridgeAddress,
                chainId,
                chainType,
                currentNonce,
                requestPayload.outboundTxNonce,
                emptyData,
                1,
                false
            );
            return;
        }

        (bool success, bytes memory data) = address(this).call(
            abi.encodeWithSelector(
                this.executeHandlerCalls.selector,
                requestPayload.routerBridgeAddress,
                requestPayload.handlers,
                requestPayload.payloads,
                requestPayload.isAtomic
            )
        );
        emit EventOutboundAck(
            requestPayload.relayerRouterAddress,
            requestPayload.routerBridgeAddress,
            chainId,
            chainType,
            currentNonce,
            requestPayload.outboundTxNonce,
            data,
            0,
            success
        );
    }

    function executeHandlerCalls(
        string memory sender,
        bytes[] memory handlers,
        bytes[] memory payloads,
        bool isAtomic
    ) external isSelf returns (bool[] memory, bytes[] memory) {
        bool[] memory execFlags = new bool[](handlers.length);
        bytes[] memory execData = new bytes[](handlers.length);
        for (uint8 i = 0; i < handlers.length; i++) {
            address handler = toAddress(handlers[i]);
            bool success;
            bytes memory data;
            success = isContract(handler);
            if (success) {
                (success, data) = handler.call(
                    abi.encodeWithSelector(IApplication.handleRequestFromRouter.selector, sender, payloads[i])
                );
                // console.logBytes(data);
            }
            execFlags[i] = success;
            execData[i] = data;
            if (isAtomic && !success) {
                revert ExecuteCallsRevert({ execFlags: execFlags, execData: execData });
            }
        }
        return (execFlags, execData);
    }

    function requestFromSource(
        // The validators that approve the call
        Utils.ValsetArgs calldata _currentValset,
        // These are arrays of the parts of the validators signatures
        Utils.Signature[] calldata _sigs,
        Utils.CrossTalkPayload memory crossTalkPayload
    ) external nonReentrant whenNotPaused {
        // chainId, chainType data is present inside the signature validation
        // valsetArgs validation will be handled by the makeCheckPoint function
        require(
            crossTalkNonceMapping[crossTalkPayload.sourceParams.chainId][crossTalkPayload.sourceParams.chainType][
                crossTalkPayload.crossTalkNonce
            ] == 0,
            "cross-talk request message already handled"
        );

        // this will validate the checkpoint, thus verify the signers
        if (makeCheckpoint(_currentValset, chainId) != stateLastValsetCheckpoint) {
            revert Utils.IncorrectCheckpoint();
        }

        bytes memory encodedABI = abi.encode(
            // bytes32 encoding of "requestFromSource"
            REQUEST_FROM_SOURCE_METHOD_NAME,
            crossTalkPayload.eventIdentifier,
            crossTalkPayload.crossTalkNonce,
            chainType,
            chainId,
            crossTalkPayload.sourceParams.chainId,
            crossTalkPayload.sourceParams.chainType,
            crossTalkPayload.sourceParams.caller,
            crossTalkPayload.isAtomic,
            crossTalkPayload.expTimestamp,
            crossTalkPayload.contractCalls.destContractAddresses,
            crossTalkPayload.contractCalls.payloads
        );
        bytes32 messagehash = keccak256(encodedABI);
        bytes32 digest = _make_digest(messagehash);
        SignatureUtils.checkValidatorSignatures(_currentValset, _sigs, digest, Utils.constantPowerThreshold);

        crossTalkNonceMapping[crossTalkPayload.sourceParams.chainId][crossTalkPayload.sourceParams.chainType][
            crossTalkPayload.crossTalkNonce
        ] = 1;
        eventNonce++;
        uint64 currentNonce = eventNonce;
        destChainNonces[crossTalkPayload.sourceParams.chainId][crossTalkPayload.sourceParams.chainType]++;
        uint64 crossTalkNonce = destChainNonces[crossTalkPayload.sourceParams.chainId][
            crossTalkPayload.sourceParams.chainType
        ];

        if (block.timestamp > crossTalkPayload.expTimestamp) {
            bytes memory emptyData = "";
            emit CrossTalkAckEvent(
                crossTalkPayload.relayerRouterAddress,
                crossTalkPayload.sourceParams.caller,
                crossTalkPayload.sourceParams.chainId,
                crossTalkPayload.sourceParams.chainType,
                chainId,
                chainType,
                crossTalkPayload.eventIdentifier,
                crossTalkNonce,
                currentNonce,
                emptyData,
                1,
                false
            );
            return;
        }
        (bool success, bytes memory data) = address(this).call(
            abi.encodeWithSelector(
                this.executeCrossTalkCalls.selector,
                crossTalkPayload.sourceParams.caller,
                crossTalkPayload.contractCalls.destContractAddresses,
                crossTalkPayload.contractCalls.payloads,
                crossTalkPayload.isAtomic,
                crossTalkPayload.sourceParams.chainId,
                crossTalkPayload.sourceParams.chainType,
                crossTalkPayload.isReadCall
            )
        );
        emit CrossTalkAckEvent(
            crossTalkPayload.relayerRouterAddress,
            crossTalkPayload.sourceParams.caller,
            crossTalkPayload.sourceParams.chainId,
            crossTalkPayload.sourceParams.chainType,
            chainId,
            chainType,
            crossTalkPayload.eventIdentifier,
            crossTalkNonce,
            currentNonce,
            data,
            0,
            success
        );
    }

    function executeCrossTalkCalls(
        bytes memory srcContractAddress,
        bytes[] memory handlers,
        bytes[] memory payloads,
        bool isAtomic,
        string memory srcChainId,
        uint64 srcChainType,
        bool isReadCall
    ) external isSelf returns (bool[] memory, bytes[] memory) {
        bool[] memory execFlags = new bool[](handlers.length);
        bytes[] memory execData = new bytes[](handlers.length);

        for (uint8 i = 0; i < handlers.length; i++) {
            address handler = toAddress(handlers[i]);
            execFlags[i] = isContract(handler);

            if (execFlags[i]) {
                if (!isReadCall) {
                    (execFlags[i], execData[i]) = handler.call(
                        abi.encodeWithSelector(
                            ICrossTalkApplication.handleRequestFromSource.selector,
                            srcContractAddress,
                            payloads[i],
                            srcChainId,
                            srcChainType
                        )
                    );
                    // console.logBytes(execData[i]);
                } else {
                    (execFlags[i], execData[i]) = handler.call(payloads[i]);
                }
            }
            // execFlags[i] = success;
            // execData[i] = data;
            if (isAtomic && !execFlags[i]) {
                revert ExecuteCallsRevert({ execFlags: execFlags, execData: execData });
            }
        }
        return (execFlags, execData);
    }

    function crossTalkAck(
        // The validators that approve the call
        Utils.ValsetArgs calldata _currentValset,
        // These are arrays of the parts of the validators signatures
        Utils.Signature[] calldata _sigs,
        Utils.CrossTalkAckPayload memory crossTalkAckPayload
    ) external nonReentrant whenNotPaused {
        // chainId, chainType data is present inside the signature validation
        // valsetArgs validation will be handled by the makeCheckPoint function
        require(
            crossTalkNonceMapping[crossTalkAckPayload.destChainId][crossTalkAckPayload.destChainType][
                crossTalkAckPayload.crossTalkNonce
            ] == 0,
            "cross-talk request message already handled"
        );

        // this will validate the checkpoint, thus verify the signers
        if (makeCheckpoint(_currentValset, chainId) != stateLastValsetCheckpoint) {
            revert("Utils.IncorrectCheckpoint()");
        }

        bytes memory encodedABI = abi.encode(
            // bytes32 encoding of "crossTalkFromRouter"
            CROSS_TALK_ACK_METHOD_NAME,
            crossTalkAckPayload.eventIdentifier,
            crossTalkAckPayload.crossTalkNonce,
            chainType,
            chainId,
            crossTalkAckPayload.destChainType,
            crossTalkAckPayload.destChainId,
            crossTalkAckPayload.srcContractAddress,
            crossTalkAckPayload.execFlags,
            crossTalkAckPayload.execData
        );

        bytes32 messagehash = keccak256(encodedABI);

        bytes32 digest = _make_digest(messagehash);
        SignatureUtils.checkValidatorSignatures(_currentValset, _sigs, digest, Utils.constantPowerThreshold);

        crossTalkNonceMapping[crossTalkAckPayload.destChainId][crossTalkAckPayload.destChainType][
            crossTalkAckPayload.crossTalkNonce
        ] = 1;
        eventNonce++;
        (bool success, bytes memory data) = address(this).call(
            abi.encodeWithSelector(
                this.executeCrossTalkAck.selector,
                crossTalkAckPayload.eventIdentifier,
                crossTalkAckPayload.srcContractAddress,
                crossTalkAckPayload.execFlags,
                crossTalkAckPayload.execData
            )
        );
        emit CrossTalkAckReceivedEvent(
            crossTalkAckPayload.relayerRouterAddress,
            eventNonce,
            crossTalkAckPayload.eventIdentifier,
            chainType,
            chainId,
            success,
            data
        );
    }

    function executeCrossTalkAck(
        uint64 eventIdentifier,
        bytes memory srcContractAddress,
        bool[] memory execFlags,
        bytes[] memory execData
    ) external isSelf {
        address handler = toAddress(srcContractAddress);
        bool success;
        bytes memory data;
        success = isContract(handler);
        if (success) {
            (success, data) = handler.call(
                abi.encodeWithSelector(
                    ICrossTalkApplication.handleCrossTalkAck.selector,
                    eventIdentifier,
                    execFlags,
                    execData
                )
            );
        }
    }

    function isContract(address addr) internal view returns (bool) {
        uint256 size;
        assembly {
            size := extcodesize(addr)
        }
        return size > 0;
    }

    function toAddress(bytes memory _bytes) internal pure returns (address contractAddress) {
        bytes20 srcTokenAddress;
        assembly {
            srcTokenAddress := mload(add(_bytes, 0x20))
        }
        contractAddress = address(srcTokenAddress);
    }

    function toBytes(address a) public pure returns (bytes memory b) {
        assembly {
            let m := mload(0x40)
            a := and(a, 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF)
            mstore(add(m, 20), xor(0x140000000000000000000000000000000000000000, a))
            mstore(0x40, add(m, 52))
            b := m
        }
    }

    receive() external payable {}
}
